/**
   This program evaluates (interprets) an Abstract Syntax Tree (AST)
   that is defined by the following grammar.

       Prog ::= Exp
              | '(' 'prog' Exp+ Exp ')'

        Exp ::= List
              | Symbol
              | Car
              | Cdr
              | Cons
              | Empty?
              | Fun
              | Lambda
              | Apply
              | If
              | While
              | Set
              | Var
              | Begin
              | Print
              | AExp
              | BExp
              | INTEGER
              | BOOLEAN
              | VARIABLE

       List ::= '(' 'list' Symbol* ')'   // List doesn't evaluate "symbols"

     Symbol ::= '(' 'sym' SYMBOL ')'     // Construct a symbol, which isn't evaluated

        Car ::= '(' 'car' Exp ')'

        Cdr ::= '(' 'cdr' Exp ')'

       Cons ::= '(' 'cons' Exp Exp ')'

     Empty? ::= '(' 'empty?' Exp ')'

        Fun ::= '(' 'fun' VARIABLE Lambda ')'   // a function declaration

     Lambda ::= '(' 'lambda' VARIABLE* Exp ')'  // formal parameters followed by function body

      Apply ::= '(' 'apply' Exp Exp* ')'        // function value followed by actual parameters

         If ::= '(' 'if' Exp Exp Exp ')'

      While ::= '(' 'while' Exp Exp ')'

        Set ::= '(' 'set' VARIABLE Exp ')'

        Var ::= '(' 'var' VARIABLE Exp ')'

      Begin ::= '(' 'begin' Exp+ Exp ')'

      Print ::= '(' 'print' Exp ')'

       BExp ::= '(' '||'  Exp Exp+ ')'
              | '(' '&&'  Exp Exp+ ')'
              | '(' '!'   Exp ')'
              | '(' RelOp Exp Exp ')'
              | '('  EqOp Exp Exp ')'

      RelOp ::= '<' | '>' | '<=' | '>='
       EqOp ::= '==' | '!='

       AExp ::= '(' '+' Exp Exp* ')'
              | '(' '-' Exp Exp? ')'
              | '(' '*' Exp Exp+ ')'
              | '(' '/' Exp Exp  ')'
              | '(' '%' Exp Exp  ')'
              | '(' '^' Exp Exp  ')'

    INTEGER ::= [0-9]+
    BOOLEAN ::= 'true' | 'false'
   VARIABLE ::= [a-zA-Z][a-zA-Z0-9]*
     SYMBOL ::= [a-zA-Z][a-zA-Z0-9]*
*/

public class Evaluate
{
   private static final int DEBUG = 0;

   private static Environment globalEnv;  // reference to the global environment object
   private static Environment env;        // reference to the environment chain

   public static Value eval(Tree tree) throws EvalException
   {
      return evaluateProg( tree );
   }//eval()


   // Evaluate a prog
   private static Value evaluateProg(Tree tree) throws EvalException
   {
      Value result = null;

      // Instantiate the global environment
      // (it will always be at the end of the
      // environment chain).
      globalEnv = new Environment();
      env = globalEnv;

      // Check whick kind of Prog we have.
      if ( ! tree.getElement().equals("prog") )
      {
         // Evaluate the single expression.
         result = evaluateExp( tree );
      }
      else
      {
         // Evaluate each Exp in the Prog.
         // Any Fun will have the side effect of putting
         // a function name in the global environment.
         // Any Var expressions will have the side effect
         // of putting a variable in the environment chain.
         // Any Set expressions will have the side effect
         // of changing a value in the environment chain.
         // Any Print expressions will have the side effect
         // of printing an output.
         // Any other expressions would be pointless!
         for (int i = 0; i < tree.degree()-1; i++)
         {
            evaluateExp( tree.getSubTree(i) );
         }

         // Evaluate the last expression and use its
         // value as the value of the prog expression.
         result = evaluateExp( tree.getSubTree(tree.degree()-1) );
      }

      return result;
   }//evaluateProg()


   // Evaluate an expression
   public static Value evaluateExp(Tree tree) throws EvalException
   {
      Value result = null;

      String node = tree.getElement();

      if ( node.equals("list") )
         result = evaluateList( tree );
      else if ( node.equals("sym") )
         result = evaluateSymbol( tree );
      else if ( node.equals("car")
             || node.equals("cdr")
             || node.equals("cons")
             || node.equals("empty?"))
         result = evaluateListOp( tree );
      else if ( node.equals("fun") )
         result = evaluateFun( tree );
      else if ( node.equals("lambda") )
         result = evaluateLambda( tree );
      else if ( node.equals("apply") )
         result = evaluateApply( tree );
      else if ( node.equals("if") )
         result = evaluateIf( tree );
      else if ( node.equals("while") )
         result = evaluateWhile( tree );
      else if ( node.equals("set") )
         result = evaluateSet( tree );
      else if ( node.equals("var") )
         result = evaluateVar( tree );
      else if ( node.equals("begin") )
         result = evaluateBegin( tree );
      else if ( node.equals("print") )
         result = evaluatePrint( tree );
      else if ( node.equals("&&") || node.equals("||")
             || node.equals("!") )
         result = evaluateBexp(tree);  // boolean expression
      else if ( node.equals("<")  || node.equals(">")
             || node.equals("<=") || node.equals(">=")
             || node.equals("==") || node.equals("!=") )
         result = evaluateRexp(tree);  // relational operator
      else if ( node.equals("+") || node.equals("-")
             || node.equals("*") || node.equals("/")
             || node.equals("%") || node.equals("^") )
         result = evaluateAexp(tree);  // arithmetic expression
      else if ( tree.degree() == 0 )
      {
         if ( node.equals("true") || node.equals("false") )
            result = new Value( node.equals("true") );
         else if ( node.matches("^[-]*[0-9][0-9]*") )
            result = new Value( Integer.parseInt( node ) );
         else if ( env.defined(node) )  // a variable
            result = env.lookUp( node );
         else  // runtime check
            throw new EvalException("undefined variable: " + node + "\n");
      }
      else
         throw new EvalException("invalid expression: " + tree + "\n");

      return result;
   }//evaluateExp()


   // Evaluate a list literal
   private static Value evaluateList(Tree tree) throws EvalException
   {
      Value result = new Value();  // create an empty list

      // fill up the list (notice that we work from right to left in the list)
      for (int i = tree.degree()-1; i >= 0; i--)
      {
         String element = tree.getSubTree(i).getElement();

         if ( tree.getSubTree(i).degree() != 0 || element.equals("list") )
         {
            result.valueCC = new ConsCell( evaluateExp(tree.getSubTree(i)), result.valueCC );
         }
         else
         {
            Value value;

            if ( element.matches("^[0-9][0-9]*") )
            {
               value = new Value( Integer.parseInt(element) );
            }
            else if ( element.equals("true") || element.equals("false") )
            {
               value = new Value( element.equals("true") );
            }
            else if ( env.defined(element) )
            {
               value = env.lookUp(element);
            }
            else  // unevaluated string (a "symbol")
            {
               value = new Value( element );
            }

            result.valueCC = new ConsCell(value, result.valueCC);
         }
      }
      return result;
   }//evaluateList()


   // Evaluate a symbol literal
   private static Value evaluateSymbol(Tree tree)  throws EvalException
   {
      return new Value( tree.getSubTree(0).getElement() );
   }//evaluateSymbol()


   /**
      Apply a list operation to the actual parameters
      contained in the child nodes of the tree.
   */
   private static Value evaluateListOp(Tree tree) throws EvalException
   {
      Value result = null;

      String op = tree.getElement();

      if ( op.equals("car") )
      {
         Value list = evaluateExp( tree.getSubTree(0) );
         if ( ! list.tag.equals(Value.LIST_TAG) )
         {
            throw new EvalException("car: argument is not a list: " + list);
         }
         if ( list.valueCC == null )
         {
            throw new EvalException("car: argument is an empty list: " + list);
         }
         else
         {
            result = list.valueCC.car();
         }
      }
      else if ( op.equals("cdr") )
      {
         Value list = evaluateExp( tree.getSubTree(0) );
         if ( ! list.tag.equals(Value.LIST_TAG) )
         {
            throw new EvalException("cdr: argument is not a list: " + list);
         }
         if ( list.valueCC == null )
         {
            throw new EvalException("cdr: argument is an empty list: " + list);
         }
         else
         {
            result = new Value( list.valueCC.cdr() );
         }
      }
      else if ( op.equals("cons") )
      {
         Value list = evaluateExp( tree.getSubTree(1) );
         if ( ! list.tag.equals(Value.LIST_TAG) )
         {
            throw new EvalException("cons: 2nd argument is not a list: " + list);
         }
         else
         {
            Value car = evaluateExp( tree.getSubTree(0) );
            result = new Value( new ConsCell(car, list.valueCC) );
         }
      }
      else if ( op.equals("empty?") )
      {
         Value list = evaluateExp( tree.getSubTree(0) );
         if ( ! list.tag.equals(Value.LIST_TAG) )
         {
            throw new Error("empty?: argument is not a list: " + list);
         }
         else
         {
            result = new Value( list.valueCC == null );
         }
      }
      else  // shouldn't get here
      {
         throw new EvalException("illegal list function: " + op);
      }

      return result;
   }//evaluateListOp()


   /**
      Evalute a function definition expression.

      This method mutates the current local environment object.
      The Value object put into the environment by this method
      has the tag "lambda" and its value field is a reference
      to an IPEP object.

      An IPEP object is usually called a "closure" and it contains
      two references. The first reference in a closure is an IP
      (an "Instruction Pointer") and the other a reference is an EP
      (an "Environment Pointer"). The IP reference refer's to the
      function's lambda expression (its "instructions"). The EP
      reference refer's to the function's outer scope, the scope
      that is needed for looking up the function's non-local references.

      Since functions in this language can be nested, the function's
      outer scope need no longer be the global environment (as it was
      in the previous languages). The function's outer scope is whatever
      local environment is being used at the time this evaluateFun()
      is called.
   */
   public static Value evaluateFun(Tree tree) throws EvalException
   {
      Value result = null;

      // get the function name
      String name = tree.getSubTree(0).getElement();
      // check if this function has already been defined
      if ( env.definedLocal(name) )
      {
         throw new EvalException("function already exists: " + name);
      }
      // check if the definition really is a function
      if ( ! tree.getSubTree(1).getElement().equals("lambda"))
      {
         throw new EvalException("bad function definition: " + tree);
      }

      // get the function value (the closure)
      result = evaluateLambda( tree.getSubTree(1) );

      // add the <name, value> pair to the environment
      env.add(name, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateFun()


   /**
      Evalute a lambda expression.

      All of the code for this method was moved
      from the Language_8 version of evaluateFun().
   */
   public static Value evaluateLambda(Tree tree) throws EvalException /*** Language_9 change ***/
   {
      Value result = null;

      // get the lambda expression (the IP)
      Tree lambda = tree;

      // get a reference to the current local Environment object (the EP)
      Environment ep = env;

      // create an IPEP object (a closure)
      IPEP ipep = new IPEP(lambda, ep);

      // create the return value
      result = new Value(ipep);

      return result;
   }//evaluateLambda()


   /**
      This method "applies" a function value to actual parameters.
      This method evaluates the body of the function in an environment
      that binds the actual parameter values (from this function application)
      to the formal parameters (from the function's lambda expression).

      Since this language allows nested functions, the function's outer
      scope (where the function finds its non-local references) may no
      longer be the global Environment object. The function's closure
      (the IPEP object that is the functions "value") points us to the
      Environment object that the function should use for non-local
      references.
   */
   private static Value evaluateApply(Tree tree) throws EvalException
   {
      // Evaluate the apply's first parameter to a function value,
/*1*/ Value funValue = evaluateExp( tree.getSubTree(0) );
      // check that what we are applying really is a function
      if ( ! funValue.tag.equals(Value.LAMBDA_TAG))  // runtime check
      {
         throw new EvalException("bad function value: " + tree);
      }
      // get a reference to the function's  closure
      IPEP ipep = funValue.valueL;   /*** Language_8 change (1) ***/
      // and to the function's "lambda expression" and its "nesting link".
      Tree lambda    = ipep.ip;      /*** Language_8 change (2) ***/
      Environment ep = ipep.ep;      /*** Language_8 change (3) ***/

      // Check that the number of actual parameters
      // is equal to the number of formal parameters.
      // (Actually, all we really need to know is that
      // the number of actual parameters is at least
      // the number of formal parameters.)
      if ( tree.degree() != lambda.degree() )  // runtime check
      {
         throw new EvalException("wrong number of parameters: " + tree);
      }

      // Create a new environment object that is "nested"
      // in this function's outer environment (lexical scope).
      // This environment is used to bind actual parameter
      // values to formal paramter names.
/*2*/ Environment localEnv = new Environment(ep, "Function Activation");/**Language_8 change (4)**/

      // Bind, in the new environment object, the actual parameter
      // values to the formal parameter names.
/*3*/ for (int i = 1; i < tree.degree(); i++) // iterate through the actual parameters
      {
         // Evaluate, using the current environment chain
         // (NOT the new local environment object)
         // an actual parameter expression's value.
/*4*/    Value actualParamValue = evaluateExp( tree.getSubTree(i) );

         // Retrieve, from within the lambda expression,
         // a formal parameter name.
/*5*/    String formalParamName = lambda.getSubTree(i-1).getElement();

         // Bind, in the new local environment object, the actual
         // paramter value to a formal parameter name.
/*6*/    localEnv.add(formalParamName, actualParamValue);
      }

      if (DEBUG > 0) System.out.println( localEnv + "\n" ); // for debugging purposes

      // Evaluate the body of the lambda expression using the
      // new environment (which contains the binding of the actual
      // parameter values to the function's formal parameter names).
/*7*/ Environment originalEnv = env;
/*8*/ env = localEnv;
/*9*/ Value result = evaluateExp( lambda.getSubTree(tree.degree()-1) );

      // Finally, restore the environment chain.
/*10*/env = originalEnv;

      return result;
   }//evaluateApply()


   // Evaluate an if-expression
   private static Value evaluateIf(Tree tree) throws EvalException
   {
      if ( 3 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect conditional expression: " + tree + "\n");
      }

      Value result = null;

      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // do a runtime type check
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      if ( conditionalExp.valueB )
      {
         result = evaluateExp( tree.getSubTree(1) );
      }
      else
      {
         result = evaluateExp( tree.getSubTree(2) );
      }

      return result;
   }//evaluateIf()


   // Evaluate a while-loop expression
   private static Value evaluateWhile(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("incorrect while expression: " + tree + "\n");
      }

      Value result = null;

      // evaluate the boolean condition
      Value conditionalExp = evaluateExp( tree.getSubTree(0) );
      // do a runtime type check
      if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
      {
         throw new EvalException("illegal boolean expression: " + tree);
      }

      while ( conditionalExp.valueB )
      {
         // evaluate the body of the loop (for its side effects)
         evaluateExp( tree.getSubTree(1) );
         // re-evaluate the boolean condition
         conditionalExp = evaluateExp( tree.getSubTree(0) );
         // do a runtime type check
         if ( ! conditionalExp.tag.equals(Value.BOOL_TAG) )
         {
            throw new EvalException("illegal boolean expression: " + tree);
         }
      }

      // always return false for a while-loop expression
      result =  new Value( false );

      return result;
   }//evaluateWhile()


   // Evaluate a set expression
   private static Value evaluateSet(Tree tree) throws EvalException
   {
      Value result = null;

      // get the variable
      String variable = tree.getSubTree(0).getElement();
      // check if this variable has already been declared
      if ( ! env.defined(variable) )  // runtime check
      {
         throw new EvalException("undefined variable: " + variable);
      }

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      result = evaluateExp(expr);
      // update this variable in the environment
      env.update(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateSet()


   // Evaluate a var expression
   private static Value evaluateVar(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = null;

      // get the variable
      String variable = tree.getSubTree(0).getElement();

      // check if this variable has already been declared
      // in the local environment
      if ( env.definedLocal(variable) )  // runtime check
      {
         throw new EvalException("varaible already declared: " + variable + "\n");
      }

      // get, and then evaluate, the expression
      Tree expr = tree.getSubTree(1);
      result = evaluateExp( expr );

      // declare the new, local, variable
      env.add(variable, result);

      if (DEBUG > 0) System.out.println( env + "\n" ); // for debugging purposes

      return result;
   }//evaluateVar()


   // Evaluate a begin expression
   private static Value evaluateBegin(Tree tree) throws EvalException
   {
      Value result = null;

      // Create a new Environment object chained to (or "nested in")
      // the previous (:outer") environment object.
      Environment previousEnv = env;
      env = new Environment(previousEnv, "Local (begin)");

      // Evaluate each sub expression in the begin
      // expression (using the new environment chain).
      // The return value of each expression is
      // discarded, so any expression without a
      // side-effect is worthless.
      for (int i = 0; i < tree.degree()-1; i++)
      {
         evaluateExp( tree.getSubTree(i) );
      }

      // Evaluate the last expression and use its
      // value as the value of the begin expression.
      result = evaluateExp( tree.getSubTree(tree.degree()-1) );

      env = previousEnv; // Just before this method returns, we remove from the
                         // chain of Environment objects the local Environment
                         // object we created at the beginning of this method.
                         // The local Environment object becomes a garbage object,
      return result;
   }//evaluateBegin()


   // Evaluate a print expression
   private static Value evaluatePrint(Tree tree) throws EvalException
   {
      if ( 1 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      Value result = evaluateExp( tree.getSubTree(0) );

      System.out.print( result + "\n" );

      return result;
   }//evaluatePrint()


   // Evaluate a boolean expression
   private static Value evaluateBexp(Tree tree) throws EvalException
   {
      boolean result = false;

      String node = tree.getElement();

      Value value = evaluateExp( tree.getSubTree(0) );
      if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
      {
         throw new EvalException("not a boolean expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      result = value.valueB;

      if ( node.equals("&&") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result && value.valueB;
            }
            else  // short circuit the evaluation of '&&'
            {
               result = false;
               break;
            }
         }
      }
      else if ( node.equals("||") )
      {
         if ( 2 > tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         for (int i = 1; i < tree.degree(); i++)
         {
            if (! result)
            {
               value = evaluateExp( tree.getSubTree(i) );
               if ( ! value.tag.equals(Value.BOOL_TAG) )  // runtime check
               {
                  throw new EvalException("not a boolean expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result = result || value.valueB;
            }
            else  // short circuit the evaluation of '||'
            {
               result = true;
               break;
            }
         }
      }
      else if ( node.equals("!") )
      {
         if ( 1 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = ! result;
      }

      return new Value( result );
   }//evaluateBexp()


   // Evaluate a relational expression (which is a kind of boolean expression)
   private static Value evaluateRexp(Tree tree) throws EvalException
   {
      if ( 2 != tree.degree() )  // runtime check
      {
         throw new EvalException("wrong number of arguments: " + tree + "\n");
      }

      boolean result = false;

      String op = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }

      Value valueR = evaluateExp( tree.getSubTree(1) );
      if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(1) + "\n");
      }

      int resultL = valueL.valueI;
      int resultR = valueR.valueI;

      if ( op.equals("<") )
      {
         result = resultL < resultR;
      }
      else if ( op.equals(">") )
      {
         result = resultL > resultR;
      }
      else if ( op.equals("<=") )
      {
         result = resultL <= resultR;
      }
      else if ( op.equals(">=") )
      {
         result = resultL >= resultR;
      }
      else if ( op.equals("==") )
      {
         result = resultL == resultR;
      }
      else if ( op.equals("!=") )
      {
         result = resultL != resultR;
      }

      return new Value( result );
   }//evaluateRexp()


   // Evaluate an arithmetic expression
   private static Value evaluateAexp(Tree tree) throws EvalException
   {
      int result = 0;

      String node = tree.getElement();

      Value valueL = evaluateExp( tree.getSubTree(0) );
      if ( ! valueL.tag.equals(Value.INT_TAG) )  // runtime check
      {
         throw new EvalException("not a integer expression: "
                                  + tree.getSubTree(0) + "\n");
      }
      int resultL = valueL.valueI;
      int resultR = 0;

      Value valueR = null;
      if ( tree.degree() >= 2 )
      {
         valueR = evaluateExp( tree.getSubTree(1) );
         if ( ! valueR.tag.equals(Value.INT_TAG) )  // runtime check
         {
            throw new EvalException("not a integer expression: "
                                     + tree.getSubTree(1) + "\n");
         }
         resultR = valueR.valueI;
      }

      if ( node.equals("+") )
      {
         if ( tree.degree() == 1 )
            result = resultL;
         else
         {
            result = resultL + resultR;

            for (int i = 2; i < tree.degree(); i++)
            {
               Value temp = evaluateExp( tree.getSubTree(i) );
               if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
               {
                  throw new EvalException("not a integer expression: "
                                           + tree.getSubTree(i) + "\n");
               }
               result += temp.valueI;
            }
         }
      }
      else if ( node.equals("-") )
      {
         if ( 2 < tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         if ( tree.degree() == 1 )
            result = -resultL;
         else
            result = resultL - resultR;
      }
      else if ( node.equals("*") )
      {
         if ( 1 == tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }

         result = resultL * resultR;

         for (int i = 2; i < tree.degree(); i++)
         {
            Value temp = evaluateExp( tree.getSubTree(i) );
            if ( ! temp.tag.equals(Value.INT_TAG) )  // runtime check
            {
               throw new EvalException("not a integer expression: "
                                        + tree.getSubTree(i) + "\n");
            }
            result *= temp.valueI;
         }
      }
      else if ( node.equals("/") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL / resultR;
      }
      else if ( node.equals("%") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = resultL % resultR;
      }
      else if ( node.equals("^") )
      {
         if ( 2 != tree.degree() )  // runtime check
         {
            throw new EvalException("wrong number of arguments: " + tree + "\n");
         }
         result = (int)Math.pow(resultL, resultR);
      }

      return new Value( result );
   }//evaluateAexp()
}//Evaluate